import os

from django.apps import apps
from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

LINKS = None


def configure():
    global LINKS
    LINKS = getattr(settings, 'BANKLINKS', None)

    if not LINKS:
        raise ImproperlyConfigured(u"BANKLINKS not found in settings")
    elif type(LINKS) != dict:
        raise ImproperlyConfigured(u"settings.BANKLINKS must be dict")
    else:
        # Verify it contains proper data
        for bank_name, data in LINKS.items():
            if len(bank_name) > 16:
                raise ImproperlyConfigured(u"Bank's name must be at most 16 characters (%s)" % bank_name)
            if type(data) != dict:
                raise ImproperlyConfigured(u"Each bank in settings.BANKLINKS must correspond to dict with settings of that bank")

            if 'PRIVATE_KEY' not in data:
                raise ImproperlyConfigured(u"PRIVATE_KEY not found in settings for bank %s" % bank_name)
            if 'PUBLIC_KEY' not in data:
                raise ImproperlyConfigured(u"PUBLIC_KEY not found in settings for bank %s" % bank_name)
            if 'SND_ID' not in data:
                raise ImproperlyConfigured(u"SND_ID not found in settings for bank %s" % bank_name)
            if 'REQUEST_URL' not in data:
                raise ImproperlyConfigured(u"REQUEST_URL not found in settings for bank %s" % bank_name)

            if not os.path.isfile(data['PRIVATE_KEY']):
                raise ImproperlyConfigured(u"Private key file %s for bank %s does not exist." % (data['PRIVATE_KEY'], bank_name))

            if not os.path.isfile(data['PUBLIC_KEY']):
                raise ImproperlyConfigured(u"Public key file %s for bank %s does not exist." % (data['PUBLIC_KEY'], bank_name))


configure()


def get_model(model_name):
    return apps.get_model('thorbanks.%s' % model_name)


def manual_models(model_name):
    manual = getattr(settings, 'THORBANKS_MANUAL_MODELS', False)

    if manual and isinstance(manual, (list, tuple)):
        return model_name in manual

    return False


# Shorthand methods
def get_private_key(the_bank):
    return LINKS[the_bank]['PRIVATE_KEY']


def get_public_key(the_bank):
    return LINKS[the_bank]['PUBLIC_KEY']


def get_snd_id(the_bank):
    return LINKS[the_bank]['SND_ID']


def get_request_url(the_bank):
    return LINKS[the_bank]['REQUEST_URL']


def get_link_type(the_bank):
    return LINKS[the_bank]['TYPE']


def get_digest_counts_bytes(the_bank):
    """ Returns whether there should be byte- or character counts in digest used to calculate VK_MAC
        This matters when using UTF-8 multibyte characters.
        Swedbank always uses characters, SEB spec doesn't say anything but they actually expect byte-counts.
    """
    return LINKS[the_bank].get('DIGEST_COUNTS_BYTES', False)


def get_encoding(the_bank):
    """ Returns encoding to be used for the given bank.
        It defaults to UTF-8, which is fine and preferred in most cases. However, e.g. Danske only supports ISO-8859-1.
    """
    return LINKS[the_bank].get('ENCODING', 'UTF-8')


def get_bank_choices():
    """ Returns list of (bank_name, pretty_name, image_path, order) tuples.
    Useful in forms
    """
    return [(the_bank,
             LINKS[the_bank].get('PRINTABLE_NAME', the_bank),
             LINKS[the_bank].get('IMAGE_PATH', the_bank),
             LINKS[the_bank].get('ORDER', 99)) for the_bank in LINKS.iterkeys()]



import random

from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import reverse
from django.test import TestCase
from django.test.client import RequestFactory

from thorbanks import settings
from thorbanks.forms import PaymentRequest
from thorbanks.utils import pingback_url, request_digest


class TestThorBanksSettings(TestCase):
    @staticmethod
    def get_banklink_config(bank_name=None, printable_name=None):
        import os

        return {
            bank_name or 'swedbank': {
                'PRINTABLE_NAME': printable_name or 'Swedbank',
                'REQUEST_URL': 'https://pangalink.net/banklink/swedbank',
                'SND_ID': 'uid513131',
                'PRIVATE_KEY': os.path.join(os.path.dirname(__file__), '..', 'example', 'certs', 'swed_key.pem'),
                'PUBLIC_KEY': os.path.join(os.path.dirname(__file__), '..', 'example', 'certs', 'swed_cert.pem'),
                'TYPE': 'banklink',
                'IMAGE_PATH': 'swedbank.png',
                'ORDER': 1,
            },
            'seb': {
                'PRINTABLE_NAME': 'SEB',
                'REQUEST_URL': 'https://pangalink.net/banklink/seb',
                'SND_ID': 'uid513157',
                'PRIVATE_KEY': os.path.join(os.path.dirname(__file__), '..', 'example', 'certs', 'seb_key.pem'),
                'PUBLIC_KEY': os.path.join(os.path.dirname(__file__), '..', 'example', 'certs', 'seb_cert.pem'),
                'DIGEST_COUNTS_BYTES': True,
                'TYPE': 'banklink',
                'IMAGE_PATH': 'seb.png',
                'ORDER': 2,
            },
            'danske': {
                'PRINTABLE_NAME': 'Danske Bank',
                'REQUEST_URL': 'https://pangalink.net/banklink/sampo',
                'SND_ID': 'uid513160',
                'PRIVATE_KEY': os.path.join(os.path.dirname(__file__), '..', 'example', 'certs', 'danske_key.pem'),
                'PUBLIC_KEY': os.path.join(os.path.dirname(__file__), '..', 'example', 'certs', 'danske_cert.pem'),
                'ENCODING': 'ISO-8859-1',
                'TYPE': 'banklink',
                'IMAGE_PATH': 'danske.png',
                'ORDER': 3,
            },
        }

    def test_no_settings_raises_exception(self):
        with self.settings(BANKLINKS=None):
            with self.assertRaisesMessage(ImproperlyConfigured, "BANKLINKS not found in settings"):
                settings.configure()

    def test_banklinks_is_not_dict(self):
        with self.settings(BANKLINKS=True):
            with self.assertRaisesMessage(ImproperlyConfigured, "settings.BANKLINKS must be dict"):
                settings.configure()

    def test_banklink_name_too_long(self):
        conf = self.get_banklink_config(bank_name="ABCDEFGHIJKLMNOPS")

        with self.settings(BANKLINKS=conf):
            with self.assertRaisesMessage(ImproperlyConfigured, "Bank's name must be at most 16 characters (ABCDEFGHIJKLMNOPS)"):
                settings.configure()

    def test_banklink_data_not_dict(self):
        with self.settings(BANKLINKS={'swedbank': None}):
            error_msg = "Each bank in settings.BANKLINKS must correspond to dict with settings of that bank"
            with self.assertRaisesMessage(ImproperlyConfigured, error_msg):
                settings.configure()

    def test_banklink_private_key_missing(self):
        conf = self.get_banklink_config()
        del conf['swedbank']['PRIVATE_KEY']

        with self.settings(BANKLINKS=conf):
            with self.assertRaisesMessage(ImproperlyConfigured, "PRIVATE_KEY not found in settings for bank %s" % "swedbank"):
                settings.configure()

    def test_banklink_public_key_missing(self):
        conf = self.get_banklink_config()
        del conf['swedbank']['PUBLIC_KEY']

        with self.settings(BANKLINKS=conf):
            with self.assertRaisesMessage(ImproperlyConfigured, "PUBLIC_KEY not found in settings for bank %s" % "swedbank"):
                settings.configure()

    def test_banklink_snd_id_missing(self):
        conf = self.get_banklink_config()
        del conf['swedbank']['SND_ID']

        with self.settings(BANKLINKS=conf):
            with self.assertRaisesMessage(ImproperlyConfigured, "SND_ID not found in settings for bank %s" % "swedbank"):
                settings.configure()

    def test_banklink_request_url_missing(self):
        conf = self.get_banklink_config()
        del conf['swedbank']['REQUEST_URL']

        with self.settings(BANKLINKS=conf):
            with self.assertRaisesMessage(ImproperlyConfigured, "REQUEST_URL not found in settings for bank %s" % "swedbank"):
                settings.configure()

    def test_banklink_no_private_key_file(self):
        conf = self.get_banklink_config()
        conf['swedbank']['PRIVATE_KEY'] = "rand_name_%d.key" % random.randint(1, 999999)

        with self.settings(BANKLINKS=conf):
            error_msg = "Private key file %s for bank swedbank does not exist." % conf['swedbank']['PRIVATE_KEY']
            with self.assertRaisesMessage(ImproperlyConfigured, error_msg):
                settings.configure()

    def test_banklink_no_public_key_file(self):
        conf = self.get_banklink_config()
        conf['swedbank']['PUBLIC_KEY'] = "rand_name_%d.key" % random.randint(1, 999999)

        with self.settings(BANKLINKS=conf):
            error_msg = "Public key file %s for bank swedbank does not exist." % conf['swedbank']['PUBLIC_KEY']
            with self.assertRaisesMessage(ImproperlyConfigured, error_msg):
                settings.configure()

    def test_getters(self):
        conf = self.get_banklink_config()

        with self.settings(BANKLINKS=conf):
            settings.configure()

            self.assertEqual(settings.get_private_key("swedbank"), conf["swedbank"]["PRIVATE_KEY"])
            self.assertEqual(settings.get_public_key("swedbank"), conf["swedbank"]["PUBLIC_KEY"])
            self.assertEqual(settings.get_snd_id("swedbank"), conf["swedbank"]["SND_ID"])
            self.assertEqual(settings.get_request_url("swedbank"), conf["swedbank"]["REQUEST_URL"])
            self.assertEqual(settings.get_link_type("swedbank"), conf["swedbank"]["TYPE"])

            self.assertEqual(settings.get_digest_counts_bytes("swedbank"), False)
            self.assertEqual(settings.get_digest_counts_bytes("seb"), True)

            self.assertEqual(settings.get_encoding("swedbank"), 'UTF-8')
            self.assertEqual(settings.get_encoding("danske"), conf["danske"]["ENCODING"])

            the_list = settings.get_bank_choices()
            self.assertTrue(isinstance(the_list, list))

            for item in the_list:
                self.assertTrue(isinstance(item, tuple))

                self.assertEqual(len(item), 4)
                self.assertTrue(item[0] in conf.keys())
                self.assertEqual(item[1], conf[item[0]]['PRINTABLE_NAME'])
                self.assertEqual(item[2], conf[item[0]]['IMAGE_PATH'])
                self.assertEqual(item[3], conf[item[0]].get('ORDER', 99))

    def test_ensure_imports(self):
        """ This test is currently used to make coverage report correctly for all files in this project.
        """


class TestTransactionModel(TestCase):
    def test_str_method(self):
        trans = settings.get_model('Transaction')(pk=1, currency='EUR', amount=15.67, bank_name='myBank')

        self.assertEqual(trans.__str__(), "Transaction 1 - EUR 15.67 from myBank [pending]")


class TestPaymentFlow(TestCase):
    def test_pingback_with_request(self):
        factory = RequestFactory()
        request = factory.get('/test', HTTP_HOST='127.0.0.1')

        ping_back = pingback_url(request=request)

        self.assertEqual(ping_back, 'http://%s%s' % (request.META['HTTP_HOST'], reverse('thorbanks_response')))

    def test_payment_request(self):
        conf = TestThorBanksSettings.get_banklink_config()

        with self.settings(BANKLINKS=conf):
            settings.configure()

            redirect_to = "http://example.com"
            ping_back = pingback_url(base_url=redirect_to)

            self.assertEqual(ping_back, redirect_to + reverse('thorbanks_response'))

            request = PaymentRequest(
                bank_name="swedbank",
                amount=13.99,
                currency='EUR',
                redirect_to=redirect_to,
                redirect_on_failure=redirect_to,
                message="My cool payment",
                url=ping_back,
            )

            self.assertEqual(request.transaction.bank_name, "swedbank")
            self.assertEqual(request.transaction.amount, 13.99)
            self.assertEqual(request.transaction.currency, 'EUR')
            self.assertEqual(request.transaction.redirect_after_success, redirect_to)
            self.assertEqual(request.transaction.redirect_on_failure, redirect_to)
            self.assertEqual(request['VK_RETURN'].value(), ping_back)

            self.assertEqual(request_digest(request.cleaned_data, request.transaction.bank_name),
                             '0041002003008009uid513131001100513.99003EUR00213015My cool payment')

            # NOTE: This test will only work with the original certificate.
            self.assertEqual(request['VK_MAC'].value(),
                             'pFNyPhXm4xRj7RIVwTSSDCFvhtMV/VksaeMIxPbeFXXNaOZtDeKGGOMVkWDWm4alMJ7v4gX'
                             '0bVFadXeLH4FpDfaxawHcS+GLKxEh4l4QY6n+VtxVRmshymOkwrzhQYR4liYX7aS5lnVVS7'
                             'vJdzF4f9hvVEmE3d7E7Ecj9A4d5mOOGmaR3h9J1eK+jllP1p64H5s/HVoKN7pB5ITkslbHr'
                             'ybflAsPzqoCZlNiGL+N4SxIsV08MrHcKI7HgERqsOQzDyNwZHEwC9G04+nOTlwPQT/gE9gp'
                             'MDpRwVkZrbLeePn/l5o9fBVX2nqL27GOaEBUXqTKlHAX1bRiwbly6bV+vg==')



from warnings import warn

from django import forms
from django.conf import settings as django_settings
from django.forms.util import ErrorList
from django.forms.widgets import RadioFieldRenderer
from django.http import HttpResponse
from django.template.loader import render_to_string
from django.utils.encoding import force_text
from django.utils.html import conditional_escape
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext_lazy as _

from thorbanks import settings
from thorbanks.utils import create_signature
from thorbanks.signals import transaction_started
from thorbanks.utils import calculate_731_checksum


class AuthRequest(forms.Form):
    VK_SERVICE = forms.CharField(widget=forms.HiddenInput())
    VK_VERSION = forms.CharField(widget=forms.HiddenInput())
    VK_SND_ID = forms.CharField(widget=forms.HiddenInput())
    VK_REPLY = forms.CharField(widget=forms.HiddenInput())
    VK_RETURN = forms.CharField(widget=forms.HiddenInput())
    VK_DATETIME = forms.CharField(widget=forms.HiddenInput())
    VK_RID = forms.CharField(widget=forms.HiddenInput())
    VK_ENCODING = forms.CharField(widget=forms.HiddenInput())
    VK_MAC = forms.CharField(widget=forms.HiddenInput(), required=False)

    def __init__(self, bank_name, redirect_to, response_url, *args, **kwargs):
        self.auth = settings.get_model('Authentication')()
        self.auth.bank_name = bank_name
        self.auth.redirect_after_success = redirect_to
        self.auth.redirect_on_failure = redirect_to
        self.auth.save()

        initial = {
            'VK_RID': self.auth.pk,
            'VK_SERVICE': '4011',
            'VK_VERSION': '008',
            'VK_REPLY': '3012',
            'VK_SND_ID': settings.get_snd_id(bank_name),
            'VK_RETURN': response_url,
            'VK_DATETIME': self.auth.created.strftime('%Y-%m-%dT%H:%M:%S%z'),
            'VK_ENCODING': settings.get_encoding(bank_name)
        }

        super(AuthRequest, self).__init__(initial, *args, **kwargs)

        if self.is_valid():
            mac = create_signature(self.cleaned_data, self.auth.bank_name, auth=True)
            self.data['VK_MAC'] = mac
            if not self.is_valid():
                raise RuntimeError("signature is invalid")
        else:
            raise RuntimeError("invalid initial data")

    def redirect_html(self):
        """ Redirection html """
        html = '<form action="%s" method="POST" id="banklink_redirect_url" accept-charset="%s">' % (self.get_request_url(),
                                                                                                    self.get_encoding())

        for field in self:
            html += force_text(field) + "\n"

        html += '</form>'
        html += """<script type="text/javascript">
                       document.forms['banklink_redirect_url'].submit();
                   </script>
                """

        return mark_safe(html)

    def get_request_url(self):
        return settings.get_request_url(self.auth.bank_name)

    def get_encoding(self):
        return settings.get_encoding(self.auth.bank_name)


class PaymentRequest(forms.Form):
    """ Creates payment request and Transaction object.

    It also acts as a facade through which transaction object and HTTP redirect can be retrieved.
    """
    VK_SERVICE = forms.CharField(widget=forms.HiddenInput())
    VK_VERSION = forms.CharField(widget=forms.HiddenInput())
    VK_SND_ID = forms.CharField(widget=forms.HiddenInput())
    VK_STAMP = forms.CharField(widget=forms.HiddenInput())
    VK_AMOUNT = forms.CharField(widget=forms.HiddenInput())
    VK_CURR = forms.CharField(widget=forms.HiddenInput())
    VK_REF = forms.CharField(widget=forms.HiddenInput())
    VK_MSG = forms.CharField(widget=forms.HiddenInput())
    VK_MAC = forms.CharField(widget=forms.HiddenInput(), required=False)
    VK_RETURN = forms.CharField(widget=forms.HiddenInput())
    VK_LANG = forms.CharField(widget=forms.HiddenInput())
    VK_ENCODING = forms.CharField(widget=forms.HiddenInput())
    VK_CHARSET = forms.CharField(widget=forms.HiddenInput())

    def __init__(self, *args, **kwargs):
        if kwargs['bank_name'] == 'danske':
            # Note: This should also strip everything that isn't ISO-8859-1
            kwargs['message'] = kwargs['message'].replace(':', '')

        initial = {}
        self.transaction = settings.get_model('Transaction')()
        self.transaction.bank_name = kwargs['bank_name']
        self.transaction.description = kwargs['message']
        self.transaction.amount = initial['VK_AMOUNT'] = kwargs['amount']
        self.transaction.currency = initial['VK_CURR'] = kwargs['currency']
        self.transaction.message = initial['VK_MSG'] = kwargs['message']
        initial['VK_ENCODING'] = settings.get_encoding(self.transaction.bank_name)
        initial['VK_CHARSET'] = initial['VK_ENCODING']
        initial['VK_RETURN'] = kwargs.get('url')
        initial['VK_SERVICE'] = '1002'
        initial['VK_VERSION'] = '008'
        initial['VK_LANG'] = kwargs.get('language', 'EST')
        initial['VK_SND_ID'] = settings.get_snd_id(self.transaction.bank_name)
        self.transaction.redirect_after_success = kwargs['redirect_to']
        self.transaction.redirect_on_failure = kwargs['redirect_on_failure']
        self.transaction.save()

        transaction_started.send(settings.get_model('Transaction'), transaction=self.transaction)
        initial['VK_REF'] = calculate_731_checksum(self.transaction.pk)
        initial['VK_STAMP'] = self.transaction.pk

        super(PaymentRequest, self).__init__(initial, *args)

        if self.is_valid():
            mac = create_signature(self.cleaned_data, self.transaction.bank_name)
            self.data['VK_MAC'] = mac
            if not self.is_valid():
                raise RuntimeError("signature is invalid")
        else:
            raise RuntimeError("invalid initial data")

        if 'pangalink.net' in self.get_request_url():
            self.handle_exceptions()

    def handle_exceptions(self):
        """ This ensures that we send correct data for certain banks to pangalink.net which
            is more strict than the real banklink endpoints.
        """
        if self.transaction.bank_name == 'swedbank':
            # Swedbank uses VK_ENCODING instead of VK_CHARSET
            del self.fields['VK_CHARSET']
        elif self.transaction.bank_name == 'seb':
            # SEB uses VK_CHARSET instead of VK_ENCODING
            del self.fields['VK_ENCODING']
        elif self.transaction.bank_name == 'danske':
            del self.fields['VK_CHARSET']
            del self.fields['VK_ENCODING']

    def redirect_html(self):
        """ Hanzanet redirection html"""
        html = u'<form action="%s" method="POST" id="banklink_redirect_url" accept-charset="%s">' % (
            self.get_request_url(), self.get_encoding())
        for field in self:
            html += unicode(field) + u"\n"
        html += u'</form>'
        html += u'''<script type="text/javascript">
                    document.forms['banklink_redirect_url'].submit();
                    </script>'''
        return mark_safe(html)

    def submit_button(self, value=u"Make the payment"):
        html = u'<form action="%s" method="POST">' % (settings.get_request_url(self.transaction.bank_name))
        for field in self:
            html += unicode(field) + u"\n"
        html += '<input type="submit" value="%s" />' % value
        html += '</form>'
        return mark_safe(html)

    def as_html(self, with_submit=False, id="banklink_payment_form", submit_value="submit" ):
        """ return transaction form for redirect to HanzaNet """
        warn("deprecated", DeprecationWarning)
        html = u'<form action="%s" method="POST" id="%s">' % (settings.get_request_url(self.transaction.bank_name), id)
        for field in self:
            html += unicode(field) + u"\n"
        if with_submit:
            html += '<input type="submit" value="%s"/>' % (submit_value, )
        html += '</form>'
        return html

    def get_request_url(self):
        return settings.get_request_url(self.transaction.bank_name)

    def get_encoding(self):
        return settings.get_encoding(self.transaction.bank_name)

    def get_redirect_response_html(self):
        return render_to_string("thorbanks/request.html", {'form': self})

    def get_redirect_response(self):
        return HttpResponse(self.get_redirect_response_html())


class PaymentFormMixin(object):
    """
    Base form mixin that lets the user choose the bank used for payment.
    Note that you MUST add bank_name field to your form class, you can use `get_bank_name_field()` method to get a field
    object for that. E.g.

    ```
    class OrderForm(PaymentFormMixin, forms.Form):
        bank_name = PaymentFormMixin.get_bank_name_field()
        # Add your custom fields and/or inherit from ModelForm instead of Form
    ```
    """
    class BankNameFieldRenderer(RadioFieldRenderer):
        """ Augmented RadioFieldRenderer that lets us use out own CSS class.
        """
        UL_HTML = u'<ul class="payment-options clearfix">%s</ul>'
        LI_HTML = u'<li>%s</li>'

        def field_render(self, field):
            if 'id' in field.attrs:
                label_for = ' for="%s_%s"' % (field.attrs['id'], field.index)
            else:
                label_for = ''
            choice_label = conditional_escape(force_text(field.choice_label))

            return mark_safe(u'%s <label%s>%s</label>' % (field.tag(), label_for, choice_label))

        def render(self):
            return mark_safe(self.UL_HTML % '\n'.join(
                [self.LI_HTML % force_text(self.field_render(w)) for w in self]))

    def __init__(self, *args, **kwargs):
        self.banklink_order_overwrite = kwargs.pop('banklink_order_overwrite', dict())
        # None means all banks in the settings are permitted
        payment_methods = kwargs.pop('payment_methods', None)

        super(PaymentFormMixin, self).__init__(*args, **kwargs)

        self.fields['bank_name'].choices = self.get_payment_method_choices(payment_methods)
        self.fields['bank_name'].error_messages['required'] = _('You have to select a method of payment.')

    @classmethod
    def get_bank_name_field(cls):
        return forms.ChoiceField(label='', required=True,
                                 widget=forms.RadioSelect(renderer=cls.BankNameFieldRenderer))

    def get_payment_method_choices(self, all_payment_methods):
        payment_choices = []
        label_pattern = """<img alt="%s" src="%s" />"""

        for bank_id, bank_name, image_name, order in settings.get_bank_choices():
            if self.banklink_order_overwrite and bank_id in self.banklink_order_overwrite:
                order = self.banklink_order_overwrite[bank_id]

            if all_payment_methods is None or bank_id in all_payment_methods:
                label = label_pattern % (bank_name, self.get_bank_image_path(image_name))
                payment_choices.append((bank_id, mark_safe(label), order))

        return [(y[0], y[1]) for y in sorted(payment_choices, key=lambda x: x[2])]

    def get_bank_image_path(self, image_name):
        logos_dir = getattr(django_settings, 'BANKLINK_LOGO_PATH', 'img/payment/')
        return '%s%s%s' % (django_settings.STATIC_URL, logos_dir, image_name)

    def clean(self):
        _cleaned = self.cleaned_data

        if not _cleaned['bank_name']:
            self._errors['bank_name'] = self._errors.get('bank_name', ErrorList())
            self._errors['bank_name'].append(self.fields['bank_name'].error_messages['required'])

        return _cleaned

    def get_bank_id(self):
        if not self.cleaned_data:
            return None

        return self.cleaned_data['bank_name']



from django.utils.decorators import method_decorator
from django.views.generic import View
import logging

from django.http import HttpResponseRedirect
from django.http import HttpResponse, QueryDict
from django.shortcuts import get_object_or_404
from django.views.decorators.csrf import csrf_exempt

from thorbanks import settings
from thorbanks.forms import PaymentRequest, AuthRequest
from thorbanks.utils import verify_signature
from thorbanks.signals import transaction_succeeded, auth_succeeded, auth_failed
from thorbanks.signals import transaction_failed


logger = logging.getLogger(__name__)


def get_request_data(request):
    if request.method == 'POST':
        return request.POST
    else:
        return request.GET


class PaymentError(Exception):
    """Generic error class."""


class AuthError(Exception):
    """Generic error class."""


@csrf_exempt
def response(request):
    data = get_request_data(request)
    if 'VK_MAC' not in data:
        raise PaymentError("VK_MAC not in request")

    klass = settings.get_model('Transaction')
    transaction = get_object_or_404(klass, pk=data['VK_STAMP'])

    # Set proper encoding and get the data again (this time in correct encoding)
    request.encoding = settings.get_encoding(transaction.bank_name)

    # We have to manually replace POST after changing encoding because of a Django bug (?)
    request.POST = QueryDict(request.body, encoding=request.encoding)
    data = get_request_data(request)

    signature_valid = verify_signature(data, transaction.bank_name, data['VK_MAC'])
    if not signature_valid:
        raise PaymentError("Invalid signature. ")

    if data['VK_SERVICE'] == '1101':
        if transaction.status == klass.STATUS_PENDING:
            # Mark purchase as complete
            transaction.status = klass.STATUS_COMPLETED
            transaction.save()
            transaction_succeeded.send(klass, transaction=transaction)
    elif data['VK_SERVICE'] == '1901':
        if transaction.status == klass.STATUS_PENDING:
            # Mark purchase as failed
            transaction.status = klass.STATUS_FAILED
            transaction.save()
            transaction_failed.send(klass, transaction=transaction)
    else:
        logging.critical("thorbanks.views.response(): Got invalid VK_SERVICE code from bank: %s (transaction %s)",
                         data['VK_SERVICE'], data['VK_STAMP'])

        raise PaymentError("Bank sent confirmation with invalid VK_SERVICE!")

    if data['VK_AUTO'] == 'Y':
        # This is automatic pingback from the bank - send simple 200 response.
        return HttpResponse("request handled")
    else:
        # This request is from the user after being redirected from the bank to our server. Redirect her further.
        if data['VK_SERVICE'] == '1101':
            url = transaction.redirect_after_success
        else:
            url = transaction.redirect_on_failure

        return HttpResponseRedirect(url)


def create_payment_request(bank_name, message, amount, currency, pingback_url, redirect_to, redirect_on_failure):
    return PaymentRequest(
        bank_name=bank_name,
        amount=amount,
        currency=currency,
        redirect_to=redirect_to,
        redirect_on_failure=redirect_on_failure,
        message=message,
        url=pingback_url,
    )


def create_auth_request(request, bank_name, response_url, redirect_to=None):
    return AuthRequest(
        bank_name=bank_name,
        response_url=response_url,
        redirect_to=redirect_to or response_url,
    )


class AuthResponseView(View):
    def __init__(self, **kwargs):
        self.data = None
        self.url = None
        self.auth = None

        super(AuthResponseView, self).__init__(**kwargs)

    @method_decorator(csrf_exempt)
    def dispatch(self, request, *args, **kwargs):
        self.data = get_request_data(request)

        if 'VK_MAC' not in self.data:
            raise AuthError("VK_MAC not in request")

        klass = settings.get_model('Authentication')
        auth = get_object_or_404(klass, pk=self.data.get('VK_RID', None))

        # Set proper encoding and get the data again (this time in correct encoding)
        request.encoding = settings.get_encoding(auth.bank_name)

        # We have to manually replace POST after changing encoding because of a Django bug (?)
        request.POST = QueryDict(request.body, encoding=request.encoding)
        self.data = get_request_data(request)

        signature_valid = verify_signature(self.data, auth.bank_name, self.data['VK_MAC'], auth=True, response=True)
        if not signature_valid:
            raise AuthError("Invalid signature. ")

        self.data = self.data.dict()

        if self.data['VK_SERVICE'] == '3012':
            if auth.status == klass.STATUS_PENDING:
                # Mark auth as complete
                auth.raw_response = self.data
                auth.status = klass.STATUS_COMPLETED
                auth.save()
                auth_succeeded.send(klass, auth=auth)

        else:
            if auth.status == klass.STATUS_PENDING:
                # Mark auth as failed
                auth.raw_response = self.data
                auth.status = klass.STATUS_FAILED
                auth.save()
                auth_failed.send(klass, auth=auth)

        # This request is from the user after being redirected from the bank to our server. Redirect her further.
        if self.data['VK_SERVICE'] == '3012':
            self.url = auth.redirect_after_success
        else:
            self.url = auth.redirect_on_failure

        self.auth = auth

        return super(AuthResponseView, self).dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        return HttpResponseRedirect(self.url)

    def post(self, request, *args, **kwargs):
        return self.get(request, *args, **kwargs)
